#!/usr/bin/env python

# .-. .-')               ('-.      .-')    .-') _          .-. .-')    ('-.    .-')    .-') _             ('-.    .-')
# \  ( OO )             ( OO ).-. ( OO ). (  OO) )         \  ( OO ) _(  OO)  ( OO ). (  OO) )          _(  OO)  ( OO ).
#  ;-----.\  ,--.       / . --. /(_)---\_)/     '._         ;-----.\(,------.(_)---\_)/     '._ ,-.-') (,------.(_)---\_)
#  | .-.  |  |  |.-')   | \-.  \ /    _ | |'--...__)  .-')  | .-.  | |  .---'/    _ | |'--...__)|  |OO) |  .---'/    _ |
#  | '-' /_) |  | OO ).-'-'  |  |\  :` `. '--.  .--'_(  OO) | '-' /_)|  |    \  :` `. '--.  .--'|  |  \ |  |    \  :` `.
#  | .-. `.  |  |`-' | \| |_.'  | '..`''.)   |  |  (,------.| .-. `.(|  '--.  '..`''.)   |  |   |  |(_/(|  '--.  '..`''.)
#  | |  \  |(|  '---.'  |  .-.  |.-._)   \   |  |   '------'| |  \  ||  .--' .-._)   \   |  |  ,|  |_.' |  .--' .-._)   \
#  | '--'  / |      |   |  | |  |\       /   |  |           | '--'  /|  `---.\       /   |  | (_|  |    |  `---.\       /
#  `------'  `------'   `--' `--' `-----'    `--'           `------' `------' `-----'    `--'   `--'    `------' `-----'

from blastbesties.blastops import getPairs
from blastbesties.utils import outPathCheck, isfile, writePairs
from blastbesties._version import __version__

import argparse
import logging


def mainArgs():
    ### Argument handling.
    parser = argparse.ArgumentParser(
        description="Finds reciprocal best blast pairs from BLAST output format 6 (tabular). Where hits are sorted by query name then descending match quality."
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser.add_argument(
        "-a",
        "--blastAvB",
        required=True,
        default=None,
        help="Blast tab result file for fastaA query against fastaB subject",
    )
    parser.add_argument(
        "-b",
        "--blastBvA",
        required=True,
        default=None,
        help="Blast tab result file for fastaB query against fastaA subject",
    )
    parser.add_argument(
        "-l", "--minLen", default=1, help="Minimum length of hit to consider valid."
    )
    parser.add_argument(
        "-e",
        "--eVal",
        default=0.001,
        type=float,
        help="Maximum e-value to consider valid pair.",
    )
    parser.add_argument(
        "-s",
        "--bitScore",
        default=1.0,
        type=float,
        help="Minimum bitscore to consider valid pair.",
    )
    parser.add_argument(
        "-o",
        "--outFile",
        default=None,
        help="Write reciprocal blast pairs to this file.",
    )
    parser.add_argument(
        "-d",
        "--outDir",
        type=str,
        default=None,
        help="Directory for new sequence files to be written to.",
    )
    parser.add_argument(
        "--loglevel",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set logging level.",
    )
    # Parse arguments
    args = parser.parse_args()
    # Call main function
    return args


def main():
    # Get cmd line args
    args = mainArgs()

    # Set up logging
    numeric_level = getattr(logging, args.loglevel.upper(), None)

    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % args.loglevel)

    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=numeric_level
    )

    # Check input files exist
    isfile([args.blastAvB, args.blastBvA])

    # Check for output directories
    outFilePath = outPathCheck(args)

    # Screen BLAST input files for reciprocal pairs
    recipPairs = getPairs(
        args.blastAvB, args.blastBvA, args.minLen, args.eVal, args.bitScore
    )

    # Write reciprical best BLAST pairs to output
    writePairs(recipPairs, outFilePath)

    logging.info("Finished")



from collections import Counter
import logging


def getPairs(blastAvB, blastBvA, minLen=1, eVal=0.001, bitScore=0):
    """Reads in two blast tab output files and returns a list of pairs which are reciprocal best blast hits."""
    logging.info(f"Retriving reciprocal best hits between {blastAvB} and {blastBvA}")
    logging.info(
        f"Filtering blast hits: minLen={minLen}, eVal={eVal}, bitScore={bitScore}"
    )
    pairs = list()
    # Find best A-B pairs
    with open(blastAvB) as src:
        data = [rec.rstrip().split("\t") for rec in src]
        cleaned_data = [map(str.strip, line) for line in data]
        lastQuery = None
        for line in cleaned_data:
            # Map yield generator in Py3
            line = list(line)
            # Ignore lines begining with '#'
            if line[0][0] == "#":
                continue
            # Ignore if eVal is above threshold or bitscore is below threshold
            if float(line[10]) >= eVal or float(line[11]) <= bitScore:
                continue
            # Ignore if hit length is less than threshold
            if int(line[3]) <= minLen:
                continue
            # If first record for this query, store the query:target pair
            if line[0] != lastQuery:
                pairs.append(tuple(line[0:2]))
            # Update previous query ID
            lastQuery = line[0]
    # Append best B-A pairs as A-B
    with open(blastBvA) as src:
        data = [rec.rstrip().split("\t") for rec in src]
        cleaned_data = [map(str.strip, line) for line in data]
        lastQuery = None
        for line in cleaned_data:
            # Map yield generator in Py3
            line = list(line)
            # Ignore lines begining with '#'
            if line[0][0] == "#":
                continue
            # Ignore if eVal is above threshold or bitscore is below threshold
            if float(line[10]) >= eVal or float(line[11]) <= bitScore:
                continue
            # Ignore if hit length is less than threshold
            if int(line[3]) <= minLen:
                continue
            # If first record for this query, flip query:target pair and store as target:query.
            if line[0] != lastQuery:
                flipPair = (line[1], line[0])
                pairs.append(flipPair)
            # Update previous query ID
            lastQuery = line[0]
    # Generator which returns list items that occur more than once
    recipPairs = [k for (k, v) in Counter(pairs).items() if v > 1]
    logging.info(
        f"Found {len(recipPairs)} reciprocal best blast pairs passing match criteria."
    )
    return recipPairs



from os import makedirs
import logging
import os.path as op
import sys


def isfile(fileList: list):
    isMissing = False
    for infile in fileList:
        if not op.exists(infile):
            logging.error(f"Input file not found: {infile}")
            isMissing = True
    if isMissing:
        logging.error(f"Exiting.")
        sys.exit(1)


def outPathCheck(args):
    """Create path to output file."""
    if not args.outFile:
        baseA = op.splitext(op.basename(args.blastAvB))[0]
        baseB = op.splitext(op.basename(args.blastBvA))[0]
        outName = baseA + "_" + baseB + "_reciprocal_pairs.tab"
        # Replace any spaces with "_"
        outName = outName.replace(" ", "_")
    else:
        outName = args.outFile

    if args.outDir:
        absOutDir = op.abspath(args.outDir)
        # Create outDir if does not exist
        if not op.isdir(absOutDir):
            makedirs(absOutDir)
        outFilePath = op.join(args.outDir, outName)
    else:
        outFilePath = outName

    return outFilePath


def writePairs(recipPairs, outFilePath):
    """Write best batch pairs which occur in both BLAST tab files to output file."""
    logging.info(f"Writing pairs to file: {outFilePath}")
    # Open handle
    outFile = open(outFilePath, "w")
    # Write file header
    header = "\t".join(["#SetA", "SetB"])
    outFile.write(header + "\n")
    # Write pairs
    for a, b in recipPairs:
        abPair = "\t".join([a, b])
        outFile.write(abPair + "\n")
    # Close handle
    outFile.close()
